import json
from pathlib import Path

from fr_toolbelt.preprocessing import ( 
    process_documents, 
    )


# TEST OBJECTS AND UTILS #


TESTS_PATH = Path(__file__).parent

with open(TESTS_PATH / "test_documents.json", "r", encoding="utf-8") as f:
    TEST_DATA = json.load(f).get("results", [])


# preprocessing.documents #


def test_process_documents_all(documents = TEST_DATA):
    data = process_documents(documents)
    assert isinstance(data, list)
    assert len(data) == len(documents)


def test_process_documents_which_str1(documents = TEST_DATA, which="dockets"):
    data = process_documents(documents, which=which)
    assert isinstance(data, list)
    assert len(data) == len(documents)


def test_process_documents_which_str2(documents = TEST_DATA, which="agencies"):
    data = process_documents(documents, which=which)
    assert isinstance(data, list)
    assert len(data) == len(documents)


def test_process_documents_which_list(documents = TEST_DATA, which=["dockets", "agencies"]):
    data = process_documents(documents, which=which)
    assert isinstance(data, list)
    assert len(data) == len(documents)



from collections import Counter


class DuplicateError(Exception):
    """Encountered an error when processing duplicates.
    """
    pass


def _get_keys_as_tuple(document: dict, keys: tuple | list) -> tuple:
    return tuple((document.get(k) for k in keys))


def identify_duplicates(results: list[dict], key: str = None, keys: tuple | list = None) -> list[dict]:
    """Identify duplicates for further examination. Either one key or multiple keys can be provided.

    Args:
        results (list): List of results to check for duplicates.
        key (str, optional): Key representing the duplicated key: value pair. Defaults to None.
        keys (tuple | list, optional): Keys representing the duplicated key: value pairs. Defaults to None.

    Returns:
        list[dict]: Duplicated items from input list.
    """
    if (key is None) and (keys is not None):
        keys_gen = (_get_keys_as_tuple(r, keys) for r in results)
        c = Counter(keys_gen)
        dup_counts = [k for k, v in c.items() if v > 1]
        dup_items = [r for r in results if tuple((r.get(k) for k in keys)) in dup_counts]
    elif (key is not None) and (keys is None):
        key_gen = (r.get(key) for r in results)
        c = Counter(key_gen)
        dup_counts = [k for k, v in c.items() if v > 1]
        dup_items = [r for r in results if r.get(key) in dup_counts]
    else:
        raise ValueError("Must pass values to either 'key' or 'keys'.")
    # return output
    return dup_items


def remove_duplicates(results: list[dict], key: str = None, keys: tuple | list = None):
    """Filter out duplicates from list[dict] based on a key: value pair 
    ([source](https://www.geeksforgeeks.org/python-remove-duplicate-dictionaries-characterized-by-key/)).
    Keeps first instance of duplicated entry.

    Args:
        results (list): List of results to filter out duplicates.
        key (str, optional): Key representing the duplicated key: value pair. Defaults to None.
        keys (tuple | list, optional): Keys representing the duplicated key: value pairs. Defaults to None.

    Returns:
        tuple[list, int]: deduplicated list, number of duplicates removed
    """    
    initial_count = len(results)
    unique = set()
    res = []
    if (key is None) and (keys is not None):  # multiple keys
        for r in results:
            key_tuples = _get_keys_as_tuple(r, keys)
            # testing for already present value
            if key_tuples not in unique:
                res.append(r)
                # adding to set if new value
                unique.add(key_tuples)
    elif (key is not None) and (keys is None):  # one key
        for r in results:
            # testing for already present value
            if r.get(key) not in unique:
                res.append(r)
                # adding to set if new value
                unique.add(r[key])
    else:
        raise ValueError("Must pass values to either 'key' or 'keys'.")
    # return output
    filtered_count = len(res)
    return res, (initial_count - filtered_count)


def flag_duplicates(
        results: list[dict], 
        duplicates: list[dict] = None, 
        key: str = None, 
        keys: tuple | list = None
    ) -> list[dict]:
    """Flag duplicate documents based one or more key: value pairs.
    """
    if duplicates is None:
        duplicates = identify_duplicates(results, key=key, keys=keys)
    if (key is None) and (keys is not None):  # multiple keys
        duplicate_pairs = [_get_keys_as_tuple(doc, keys) for doc in duplicates]
        res = [{**doc, **{"duplicate": True}} if _get_keys_as_tuple(doc, keys) in duplicate_pairs else {**doc, **{"duplicate": False}} for doc in results]
    elif (key is not None) and (keys is None):  # one key
        duplicate_pairs = [doc.get(key) for doc in duplicates]
        res = [{**doc, **{"duplicate": True}} if doc.get(key) in duplicate_pairs else {**doc, **{"duplicate": False}} for doc in results]
    else:
        raise ValueError("Must pass values to either 'key' or 'keys'.")
    # return output
    return res


def process_duplicates(
        results: list[dict], 
        how: str, 
        key: str = None, 
        keys: tuple | list = None, 
        report_drop: bool = False
    ) -> list[dict]:
    """Process duplicates based on one or more key: value pairs. Options include "raise", "flag", and "drop". 
    """
    duplicates = identify_duplicates(results, key=key, keys=keys)
    count_dups = len(duplicates)
    if count_dups > 0:
        if not isinstance(how, str):
            raise TypeError
        match how:  # match options: raise, flag, drop, wildcard
            case "raise":
                raise DuplicateError(f"Results contain {count_dups} duplicate values based on {key}.")
            case "flag":
                res = flag_duplicates(results, duplicates=duplicates, key=key, keys=keys)
            case "drop":
                res, removed = remove_duplicates(results, key=key, keys=keys)
                if report_drop:
                    print(f"Removed {removed} duplicates")
            case _:
                raise ValueError("Invalid input for 'how' parameter.")
    else:
        res = results
    return res



import json
from pathlib import Path

from fr_toolbelt.preprocessing import ( 
    Presidents, 
    )


# TEST OBJECTS AND UTILS #


TESTS_PATH = Path(__file__).parent

with open(TESTS_PATH / "test_documents.json", "r", encoding="utf-8") as f:
    TEST_DATA = json.load(f).get("results", [])


# preprocessing.presidents #


def test_extract_president_info(documents = TEST_DATA):
    prez = Presidents(documents)
    data = (prez._extract_field_info(doc) for doc in prez.documents)
    assert all((isinstance(d, str) or d is None) for d in data)


def test_create_president_key(documents = TEST_DATA):
    prez = Presidents(documents)
    data = (prez._create_value_key(doc, values="test") for doc in prez.documents)
    assert all((isinstance(d, dict) and d.get(prez.value_key) == "test") for d in data)


def test_process_president_data(documents = TEST_DATA):
    prez = Presidents(documents)
    data = prez.process_data()
    assert isinstance(data, list)
    assert len(data) == len(documents)



import json
from pathlib import Path

from fr_toolbelt.utils import (
    process_duplicates, 
    DuplicateError, 
    )


TESTS_PATH = Path(__file__).parent
with open(TESTS_PATH / "test_documents.json", "r", encoding="utf-8") as f:
    TEST_DATA = json.load(f).get("results", [])


def test_process_duplicates_raise(results = TEST_DATA + TEST_DATA[0:2]):
    test_error = None
    results_out = None
    try:
        results_out = process_duplicates(results, "raise", key="document_number")
    except DuplicateError as e:
        test_error = e
        assert isinstance(e, DuplicateError)
    assert test_error is not None, f"{test_error=}"
    assert results_out is None
    

def test_process_duplicates_flag(results = TEST_DATA + TEST_DATA[0:2]):
    results_out = process_duplicates(results, "flag", key="document_number")
    flagged = [r for r in results_out if r.get("duplicate", False) == True]
    not_flagged = [r for r in results_out if r.get("duplicate", True) == False]
    #print(flagged)
    assert len(flagged) == (len(TEST_DATA[0:2]) * 2), f"No. flagged documents, {len(flagged)}, does not match duplicates."
    assert len(not_flagged) == (len(TEST_DATA) - len(TEST_DATA[0:2])), f"No. unflagged documents, {len(not_flagged)}, does not match uniques."
    assert all(1 if num in set(doc.get("document_number") for doc in flagged) else 0 for num in (doc.get("document_number") for doc in TEST_DATA[0:2]))


def test_process_duplicates_drop_key(results = TEST_DATA + TEST_DATA[0:2]):
    results_out = process_duplicates(results, "drop", key="document_number")
    assert len(results_out) == len(TEST_DATA)


def test_process_duplicates_drop_keys(results = TEST_DATA + TEST_DATA[0:2]):
    results_out = process_duplicates(results, "drop", keys=("document_number", "citation"))
    assert len(results_out) == len(TEST_DATA), f"{len(results_out)}, {len(TEST_DATA)}"


def test_process_duplicates_match_wildcard(results = TEST_DATA + TEST_DATA[0:2]):
    test_error = None
    results_out = None
    try:
        results_out = process_duplicates(results, "test", key="document_number")
    except ValueError as e:
        test_error = e
        assert isinstance(e, ValueError)
    assert test_error is not None, f"{test_error=}"
    assert results_out is None



"""
Preprocessing Federal Register API results.
"""

from .agencies import AgencyMetadata, AgencyData, INDEPENDENT_REG_AGENCIES

from .dockets import RegsDotGovData, Dockets

from .documents import process_documents

from .presidents import Presidents

from .rin import RegInfoData

__all__ = [
    "AgencyMetadata",
    "AgencyData", 
    "INDEPENDENT_REG_AGENCIES",
    "RegsDotGovData", 
    "Dockets", 
    "process_documents", 
    "Presidents", 
    "RegInfoData", 
    ]



from copy import deepcopy
from datetime import datetime, date
from pathlib import Path
import re
import time

from dateutil import tz
from pandas import DataFrame, read_csv, read_excel
import requests

from ..utils.duplicates import process_duplicates
from ..utils.format_dates import DateFormatter

# get patched version of progress bar
from ..utils.patch_progress import getpatchedprogress
progress = getpatchedprogress()
from progress.bar import Bar

BASE_PARAMS = {
    "per_page": 1000, 
    "page": 0, 
    "order": "oldest"
    }
BASE_URL = r"https://www.federalregister.gov/api/v1/documents.json?"
DEFAULT_FIELDS = (
    "document_number", 
    "citation", 
    "publication_date", 
    "agency_names", 
    "agencies", 
    "title", 
    "type", 
    "action",
    "abstract", 
    "docket_ids", 
    "dockets",
    "president",
    "regulation_id_number_info", 
    "regulations_dot_gov_info",
    "correction_of",
    "json_url",
    "html_url", 
    )

EST = tz.gettz("EST")
TODAY_EST = datetime.now(tz=EST).date()


# -- functions for handling API requests -- #


class QueryError(Exception):
    pass


class InputFileError(Exception):
    pass


def sleep_retry(timeout: int, retry: int = 3):
    """Decorator to sleep and retry request when receiving an error 
    (source: [RealPython](https://realpython.com/python-sleep/#adding-a-python-sleep-call-with-decorators)).

    Args:
        timeout (int): Number of seconds to sleep after error.
        retry (int, optional): Number of times to retry. Defaults to 3.
    """    
    def retry_decorator(function):
        def wrapper(*args, **kwargs):
            retries = 0
            while retries < retry:
                try:
                    value = function(*args, **kwargs)
                    if value is not None:
                        return value
                    else:
                        raise QueryError
                except (requests.HTTPError, requests.JSONDecodeError, ):
                    #print(f'Sleeping for {timeout} seconds')
                    time.sleep(timeout)
                    retries += 1
        return wrapper
    return retry_decorator


def _ensure_json_response(response: requests.Response):
    """Ensure request response is valid JSON by checking for 200 status code. 
    Returns JSON response or empty dictionary.
    """
    if response.status_code == 200:
        res_json = response.json()
    else:
        res_json = {}
    return res_json


@sleep_retry(60)
def _retrieve_results_by_page_range(num_pages: int, endpoint_url: str, dict_params: dict) -> list:
    """Retrieve documents by looping over a given number of pages.

    Args:
        num_pages (int): Number of pages to retrieve documents from.
        endpoint_url (str): URL for API endpoint.
        dict_params (dict): Paramters to pass in GET request.

    Returns:
        list: Documents retrieved from the API.
    """
    results, tally = [], 0
    for page in range(1, num_pages + 1):  # grab results from each page
        dict_params.update({"page": page})
        response = requests.get(endpoint_url, params=dict_params)
        response = _ensure_json_response(response)
        results_this_page = response.get("results", [])
        results.extend(results_this_page)
        tally += len(results_this_page)
    count = response.json()["count"]
    print(count, tally)
    return results, count


@sleep_retry(60)
def _retrieve_results_by_next_page(endpoint_url: str, dict_params: dict) -> list:
    """Retrieve documents by accessing "next_page_url" returned by each request.

    Args:
        endpoint_url (str): url for documents.{format} endpoint.
        dict_params (dict): Paramters to pass in GET request.

    Raises:
        QueryError: Failed to retrieve documents from all pages.

    Returns:
        list: Documents retrieved from the API.
    """
    results = []
    response = requests.get(endpoint_url, params=dict_params)
    response = _ensure_json_response(response)
    pages = response.get("total_pages", 1)
    next_page_url = response.get("next_page_url")
    counter = 0
    while next_page_url is not None:
        counter += 1
        results_this_page = response.get("results", [])
        results.extend(results_this_page)
        response = requests.get(next_page_url).json()
        next_page_url = response.get("next_page_url")
    else:
        counter += 1
        results_this_page = response.get("results", [])
        results.extend(results_this_page)
    
    # raise exception if failed to access all pages
    if counter != pages:
        raise QueryError(f"Failed to retrieve documents from {pages} pages.")
    
    return results


def _query_documents_endpoint(
        endpoint_url: str, 
        dict_params: dict, 
        handle_duplicates: bool | str = False, 
        **kwargs
    ) -> tuple[list, int]:
    """GET request for documents endpoint.

    Args:
        endpoint_url (str): URL for API endpoint.
        dict_params (dict): Paramters to pass in GET request.

    Returns:
        tuple[list, int]: Tuple of API results, count of documents retrieved.
    """    
    results, running_count = [], 0
    response = requests.get(endpoint_url, params=dict_params)
    #print(response.url)
    res_json = response.json()
    max_documents_threshold = 10000
    response_count = res_json["count"]
    
    # handles queries returning no documents
    if response_count == 0:
        pass
    
    # handles queries that need multiple requests
    elif response_count > max_documents_threshold:
        
        # get range of dates
        start_date = DateFormatter(dict_params.get("conditions[publication_date][gte]"))
        end_date = DateFormatter(dict_params.get("conditions[publication_date][lte]", f"{TODAY_EST}"))
        
        # set range of years
        start_year = start_date.year
        end_year = end_date.year
        years = range(start_year, end_year + 1)
        
        # format: YYYY-MM-DD
        quarters = ("Q1", "Q2", "Q3", "Q4")
        
        # retrieve documents
        dict_params_qrt = deepcopy(dict_params)

        with Bar(kwargs.get("message", "Years retrieved"), max=len(years)) as bar:
            for year in years:
                for quarter in quarters:
                    
                    # set start and end dates based on input date
                    gte = start_date.date_in_quarter(year, quarter, return_quarter_end=False)
                    lte = end_date.date_in_quarter(year, quarter)
                    if start_date.greater_than_date(lte):
                        # skip quarters where start_date is later than last day of quarter
                        continue
                    elif end_date.less_than_date(gte):
                        # skip quarters where end_date is ealier than first day of quarter
                        break
                    
                    # update parameters by quarter
                    dict_params_qrt.update({
                        "conditions[publication_date][gte]": f"{gte}", 
                        "conditions[publication_date][lte]": f"{lte}"
                                            })
                    
                    # get documents
                    results_qrt = _retrieve_results_by_next_page(endpoint_url, dict_params_qrt)
                    results.extend(results_qrt)
                    running_count += len(results_qrt)
                bar.next()
                
    # handles normal queries
    elif response_count in range(max_documents_threshold + 1):
        running_count += response_count
        results.extend(_retrieve_results_by_next_page(endpoint_url, dict_params))
    
    # otherwise something went wrong
    else:
        raise QueryError(f"Query returned document count of {response_count}.")
    
    if running_count != response_count:
        #print(running_count)
        #print(results[-1])
        raise QueryError(f"Failed to retrieve all {response_count} documents.")
    
    if not handle_duplicates:
        pass
    else:
        results = process_duplicates(results, how=handle_duplicates, keys=("document_number", "citation"))
    return results, running_count


# -- retrieve documents using date range -- #


def get_documents_by_date(start_date: str | date, 
                          end_date: str | date | None = None, 
                          document_types: tuple | list = None,
                          fields: tuple[str] | list[str] = DEFAULT_FIELDS,
                          endpoint_url: str = BASE_URL, 
                          dict_params: dict = BASE_PARAMS, 
                          handle_duplicates: bool | str = False, 
                          **kwargs
                          ):
    """Retrieve Federal Register documents using a date range.

    Args:
        start_date (str): Start date when documents were published (inclusive; format must be "yyyy-mm-dd").
        end_date (str, optional): End date (inclusive; format must be "yyyy-mm-dd"). Defaults to None (implies end date is today for EST timezone).
        document_types (tuple[str] | list[str], optional): If passed, only return specific document types. 
        Valid types are "RULE" (final rules), "PRORULE" (proposed rules), "NOTICE" (notices), and "PRESDOCU" (presidential documents). Defaults to None.
        fields (tuple | list, optional): Fields/columns to retrieve. Defaults to constant DEFAULT_FIELDS.
        endpoint_url (str, optional): Endpoint url. Defaults to r"https://www.federalregister.gov/api/v1/documents.json?".

    Returns:
        tuple[list, int]: Tuple of API results, count of documents retrieved.
    """
    # Not passing end_date implies end date of today EST
    if end_date is None:
        end_date = TODAY_EST

    # update dictionary of parameters
    params = dict_params.copy()
    params.update({
        "conditions[publication_date][gte]": f"{start_date}", 
        "conditions[publication_date][lte]": f"{end_date}", 
        "fields[]": fields, 
        })
    
    if document_types is not None:
        params.update({"conditions[type][]": list(document_types)})
    
    results, count = _query_documents_endpoint(
        endpoint_url, 
        params, 
        handle_duplicates=handle_duplicates, 
        **kwargs
        )
    return results, count


# -- retrieve documents using input file -- #


def get_documents_by_number(document_numbers: list, 
                            fields: tuple | list = DEFAULT_FIELDS, 
                            sort_data: bool = True
                            ):
    """Retrieve Federal Register documents using a list of document numbers.

    Args:
        document_numbers (list): Documents to retrieve based on "document_number" field.
        fields (tuple, optional): Fields/columns to retrieve. Defaults to constant DEFAULT_FIELDS.
        sort_data (bool, optional): Sort documents by "document_number". Defaults to True.

    Returns:
        tuple[list, int]: Tuple of API results, count of documents retrieved.
    """    
    if sort_data:
        document_numbers_str = ",".join(sorted(document_numbers))
    else:
        document_numbers_str = ",".join(document_numbers)
    
    endpoint_url = fr"https://www.federalregister.gov/api/v1/documents/{document_numbers_str}.json?"
    dict_params = {"fields[]": fields}
    results, count = _query_documents_endpoint(endpoint_url, dict_params)
    return results, count


def _extract_document_numbers(
        df: DataFrame, 
        pattern: str = r"(?:[a-z]\d-)?[\w|\d]{2,4}-[\d]{5,}", 
        alt_column: str = "html_url"
    ) -> list[str]:
    r"""Extract list of Federal Register document numbers from a DataFrame.

    Args:
        df (DataFrame): Input data.
        pattern (str, optional): Regex pattern for identifying document numbers from url. 
        Defaults to r"(?:[a-z]\d-)?[\w|\d]{2,4}-[\d]{5,}".
        alt_column (str, optional): Alternate column to search for document numbers. Defaults to "html_url".

    Returns:
        list[str]: List of document numbers.
    """
    if "document_number" in df.columns:
        document_numbers = [f"{doc}".strip() for doc in df.loc[:, "document_number"].to_numpy()]
    else:
        url_list = df.loc[:, alt_column].to_numpy()
        document_numbers = [re.search(pattern, url).group(0) for url in url_list]
    
    return document_numbers


def parse_document_numbers(path: Path, alt_column: str = "html_url"):
    """Parse Federal Register document numbers from input data file.

    Args:
        path (Path): Path to input file.
        alt_column (str, optional): Alternate column to search for document numbers. Defaults to "html_url".

    Raises:
        InputFileError: Raises error when input file is not in right format or does not exist.

    Returns:
        list: List of document numbers.
    """    
    try:
        file = next(p for p in path.iterdir() if (p.is_file() and p.name != ".gitignore"))
    except StopIteration as err:
        print(f"Handled exception: {err}")
        raise InputFileError("Missing input file with document numbers.")
    
    if file.suffix in (".csv", ".txt", ".tsv"):
        with open(file, "r") as f:
            df = read_csv(f)
    elif file.suffix in (".xlsx", ".xls", ".xlsm"):
        with open(file, "rb") as f:
            df = read_excel(f)
    else:
        raise InputFileError("Input file must be in CSV or Excel spreadsheet format.")
    
    return _extract_document_numbers(df, alt_column=alt_column)



from datetime import date
from platform import python_version_tuple
import re
from fr_toolbelt.utils import DateFormatter


def test__convert_to_datetime_date(
        success = ("2024-01-01", "20240101", "2024-W01-1", date(2024, 1, 1))
    ):
    
    for attempt in success:
        if (re.fullmatch(r"\d{4}-\d{2}-\d{2}", f"{attempt}", flags=re.I) is not None) or (int(python_version_tuple()[1]) >= 11):
            print(attempt)
            fdate = DateFormatter(attempt)
            result = fdate._DateFormatter__convert_to_datetime_date(attempt)
            assert isinstance(result, date)
        else:
            continue


def test_property_year(
        input_success: dict = {"string": "2023-01-01", "year": 2023}, 
        input_fail: str = "01/01/2023"
    ):
    
    fdate = DateFormatter(input_success.get("string"))
    year = fdate.year
    assert isinstance(year, int)
    assert year == input_success.get("year")
    
    try:
        DateFormatter(input_fail).year
    except ValueError as e:
        assert e.__class__ == ValueError


def test_property_formatted_date(
        success = ("2024-01-01", "20240101", "2024-W01-1", date(2024, 1, 1))
    ):
    for attempt in success:
        if (re.fullmatch(r"\d{4}-\d{2}-\d{2}", f"{attempt}", flags=re.I) is not None) or (int(python_version_tuple()[1]) >= 11):
            result = DateFormatter(attempt).formatted_date
            assert isinstance(result, date)
        else:
            try:
                result = DateFormatter(attempt).formatted_date
            except ValueError as err:
                assert isinstance(err, ValueError), f"{err=}"


def test_date_in_quarter():
    
    attempt1, attempt2, year, quarter = date(2023, 1, 1), date(2023, 4, 1), "2023", "Q1"
    
    fdate1 = DateFormatter(attempt1)
    result = fdate1.date_in_quarter(year, quarter)
    assert attempt1 == result
    
    fdate2 = DateFormatter(attempt2)
    result = fdate2.date_in_quarter(year, quarter)
    assert attempt2 != result
    assert f"{result}" == f"{year}-{fdate2.quarter_schema.get(quarter)[1]}", "should return end of Q1"
    
    result = fdate2.date_in_quarter(year, quarter, return_quarter_end=False)
    assert attempt2 != result
    assert f"{result}" == f"{year}-{fdate2.quarter_schema.get(quarter)[0]}", "should return beginning of Q1"


def test_gt_date_exclusive(inputs = {
        "earlier": date(2023, 1, 1), 
        "later": date(2023, 4, 1)}
    ):
    fdate = DateFormatter(inputs.get("later"))
    res = fdate.greater_than_date(inputs.get("earlier"))
    assert res, "later > earlier"
    res2 = fdate.greater_than_date(inputs.get("later"))
    assert not res2, "later !> later (exclusive)"
    
    fdate = DateFormatter(inputs.get("earlier"))
    res = fdate.greater_than_date(inputs.get("later"))
    assert not res, "earlier !> later"
    res2 = fdate.greater_than_date(inputs.get("earlier"))
    assert not res2, "earlier !> earlier (exclusive)"


def test_gt_date_inclusive(inputs = {
        "earlier": date(2023, 1, 1), 
        "later": date(2023, 4, 1)
        }
    ):
    fdate = DateFormatter(inputs.get("later"))
    res = fdate.greater_than_date(inputs.get("later"), inclusive=True)
    assert res, "later == later (inclusive)"
    
    fdate = DateFormatter(inputs.get("earlier"))
    res = fdate.greater_than_date(inputs.get("earlier"), inclusive=True)
    assert res, "earlier == earlier (inclusive)"


def test_lt_date_exclusive(inputs = {
        "earlier": date(2023, 1, 1), 
        "later": date(2023, 4, 1)
        }
    ):
    fdate = DateFormatter(inputs.get("later"))
    res = fdate.less_than_date(inputs.get("earlier"))
    assert not res, "later !< earlier"
    res2 = fdate.less_than_date(inputs.get("later"))
    assert not res2, "later !< later (exclusive)"
    
    fdate = DateFormatter(inputs.get("earlier"))
    res = fdate.less_than_date(inputs.get("later"))
    assert res, "earlier < later"
    res2 = fdate.less_than_date(inputs.get("earlier"))
    assert not res2, "earlier !< earlier (exclusive)"


def test_lt_date_inclusive(inputs = {
        "earlier": date(2023, 1, 1), 
        "later": date(2023, 4, 1)
        }
    ):
    fdate = DateFormatter(inputs.get("later"))
    res = fdate.less_than_date(inputs.get("later"), inclusive=True)
    assert res, "later == later (inclusive)"
    
    fdate = DateFormatter(inputs.get("earlier"))
    res = fdate.less_than_date(inputs.get("earlier"), inclusive=True)
    assert res, "earlier == earlier (inclusive)"



from .fields import FieldData


class RegInfoData(FieldData):
    """Class for processing Regulation Identifier Number (RIN) data (sourced from [RegInfo](https://www.reginfo.gov/public/jsp/Utilities/faq.jsp#dashboard)).
    Inherits from `FieldData`.
    """
    def __init__(self, 
                 documents: list[dict], 
                 field_key: str = "regulation_id_number_info", 
                 subfield_keys: tuple[str] = ("priority_category", "issue"), 
                 value_keys: tuple[str] = ("rin", "rin_priority")
                 ) -> None:
        super().__init__(documents=documents, field_key=field_key, subfield_keys=subfield_keys, value_keys=value_keys)

    def _extract_field_info(self, document: dict) -> tuple:
        
        field_info = document.get(self.field_key, {})
        
        tuple_list = []
        if len(field_info)==0:
            tuple_list.append(None)
        else:
            for k, v in field_info.items():
                if v:
                    n_tuple = (k, v.get(self.subfield_keys[0]), v.get(self.subfield_keys[1]))
                else:
                    n_tuple = (k, "", "")
                    
            tuple_list.append(n_tuple)
            try:
                tuple_list.sort(reverse=True, key=lambda x: x[2])
            except IndexError:
                pass
        
        # only return RIN info from most recent Unified Agenda issue
        return tuple_list[0]



def main():
    from pprint import pprint

    from fr_toolbelt.api_requests import (
        get_documents_by_date, 
    )

    from fr_toolbelt.preprocessing import (
        process_documents, 
        AgencyMetadata, 
        AgencyData, 
        Dockets, 
        Presidents, 
        RegInfoData, 
    )
    
    start = "2024-01-01"
    end = "2024-01-31"
    results, count = get_documents_by_date(start, end)
    print(f"Retrieved {count} documents published from {start} to {end}.")
    print("\n\n-- PRINTING AN ILLUSTRATIVE DOCUMENT BEFORE PROCESSING --\n")
    pprint(results[0])
    
    metadata, schema = AgencyMetadata().get_agency_metadata()
    agency_data = AgencyData(results, metadata, schema)
    processed_agencies = agency_data.process_data(return_format="name")
    print("\n\n-- AFTER PROCESSING AGENCIES --\n")
    pprint(processed_agencies[0])
    
    dockets = Dockets(processed_agencies)
    processed_dockets = dockets.process_data(del_keys="docket_ids")
    print("\n\n-- AFTER PROCESSING DOCKETS (REGULATIONS.GOV INFO) --\n")
    pprint(processed_dockets[0])
    
    presidents = Presidents(processed_dockets)
    processed_presidents = presidents.process_data()
    print("\n\n-- AFTER PROCESSING PRESIDENTS --\n")
    pprint(processed_presidents[0])
    
    rin = RegInfoData(processed_presidents)
    processed_rin = rin.process_data()
    print("\n\n-- AFTER PROCESSING RIN (REGINFO.GOV INFO) --\n")
    pprint(processed_rin[0])
    
    processed_docs = process_documents(results)
    print("\n\n-- AFTER PROCESSING ALL FIELDS --\n")
    pprint(processed_docs[0])

    del_keys = "type"
    processed_docs = process_documents(results, del_keys=del_keys)
    print(f"\n\n-- AFTER PROCESSING ALL FIELDS AND DELETING 1 FIELD ({del_keys}) --\n")
    pprint(processed_docs[0])

    del_keys = ("type", "docket_ids")
    processed_docs = process_documents(results, del_keys=del_keys, return_values_as_str=False, identify_ira=False)
    print(f"\n\n-- AFTER PROCESSING ALL FIELDS, DELETING 2 FIELDS ({'; '.join(del_keys)}), RETURNING AGENCY INFO AS LISTS, AND NOT IDENTIFYING INDEPENDENT REG AGENCIES --\n")
    pprint(processed_docs[0])


if __name__ == "__main__":
    print("Executing this package as a module from the command line provides a quick way to test its functionality.")
    proceed = input("Would you like to proceed? [yes/no] >>> ").lower()
    while True:
        if proceed in ("yes", "y"):
            main()
            break
        elif proceed in ("no", "n"):
            break
        else:
            proceed = input("Invalid input.\nWould you like to proceed? [yes/no] >>> ").lower()
            continue



"""
fr-toolbelt

Toolbelt of classes and functions written in Python to use with the Federal Register (FR) API.
"""

from fr_toolbelt import api_requests
from fr_toolbelt import preprocessing
from fr_toolbelt import utils

__title__ = "fr-toolbelt"
__version__ = "0.1.4"

__all__ = [
    "api_requests", 
    "preprocessing", 
    "utils", 
    "__title__", 
    "__version__",
]



from datetime import date
from platform import python_version_tuple
import re


class DateFormatError(Exception):
    pass


class DateFormatter:
    
    def __init__(self, input_date: date | str) -> None:
        self.input_date = self.__val_isoformat(input_date)
        self._formatted_date: date = self.__convert_to_datetime_date(self.input_date)
        self._year: int | None = self._formatted_date.year
        self.quarter_schema = {
            "Q1": ("01-01", "03-31"), 
            "Q2": ("04-01", "06-30"), 
            "Q3": ("07-01", "09-30"), 
            "Q4": ("10-01", "12-31"), 
            }
    
    def __val_isoformat(self, input_date: str | date):
        if isinstance(input_date, date):
            pass
        elif isinstance(input_date, str):        
            if (re.fullmatch(r"\d{4}-\d{2}-\d{2}", f"{input_date}", flags=re.I) is not None) or (int(python_version_tuple()[1]) >= 11):
                pass
            elif re.fullmatch(r"\d{8}", f"{input_date}", flags=re.I) is not None:
                input_date = f"{input_date[0:4]}-{input_date[4:6]}-{input_date[6:8]}"
            else:
                raise ValueError(f"Inappropriate argument value {input_date} for parameter 'input_date'. For more info, see the 'datetime' module docs.")
        else:
            raise TypeError(f"Inappropriate argument type {type(input_date)} for parameter 'input_date'.")
        return input_date

    def __convert_to_datetime_date(self, input_date: date | str) -> date:
        """Converts `self.input_date` from `str` to `datetime.date`. Returns input if already in proper format.
        Input string should be in any valid ISO 8601 format (e.g., for Jan. 1, 2024 -> 2024-01-01, 20240101, 2024-W01-1).

        Raises:
            TypeError: Inappropriate argument type for input_date parameter.

        Returns:
            date: Date object.
        """    
        if isinstance(input_date, date):
            return input_date
        elif isinstance(input_date, str):
            return date.fromisoformat(input_date)
        else:
            raise TypeError(f"Inappropriate argument type {type(input_date)} for parameter 'input_date'.")
    
    @property
    def formatted_date(self) -> date:
        """Formatted `datetime.date` object derived from the input value.
        """
        return self._formatted_date

    @property
    def year(self) -> int:
        """`year` instance attribute of the formatted date.
        """
        return self._year
    
    def date_in_quarter(self, check_year: str, check_quarter: str, return_quarter_end: bool = True) -> date:
        """Checks if given date falls within a year's quarter. 
        Returns input date if True, otherwise returns first or last day of quarter.

        Args:
            check_year (str): Year to check against.
            check_quarter (str): Quarter to check against.
            return_quarter_end (bool, optional): Return end date of quarter when input not in range. Defaults to True.

        Raises:
            TypeError: Inappropriate argument type for 'input_date'.

        Returns:
            datetime.date: Returns input_date when it falls within range; otherwise returns boundary date of quarter.
        """
        quarter_range = self.quarter_schema.get(f"{check_quarter}".upper())
        check_date = self.__convert_to_datetime_date(self.input_date)
        range_start = date.fromisoformat(f"{check_year}-{quarter_range[0]}")
        range_end = date.fromisoformat(f"{check_year}-{quarter_range[1]}")
        in_range = (check_date >= range_start) and (check_date <= range_end)
        #return in_range
        if in_range:
            return check_date
        elif (not in_range) and return_quarter_end:
            return range_end
        elif (not in_range) and (not return_quarter_end):
            return range_start
        else:
            raise DateFormatError

    def greater_than_date(self, comparison_date: date | str, inclusive: bool = False) -> bool:
        """Compare whether a formatted date occurs after a given comparison date.

        Args:
            comparison_date (date | str): A given date to compare with the formatted date.

        Returns:
            bool: True if formatted date occurs after comparison date.
        """    
        if inclusive:
            return self.formatted_date >= self.__convert_to_datetime_date(comparison_date)
        else:
            return self.formatted_date > self.__convert_to_datetime_date(comparison_date)
    
    def less_than_date(self, comparison_date: date | str, inclusive: bool = False) -> bool:
        """Compare whether a formatted date occurs after a given comparison date.

        Args:
            comparison_date (date | str): A given date to compare with the formatted date.

        Returns:
            bool: True if formatted date occurs before comparison date.
        """
        if inclusive:
            return self.formatted_date <= self.__convert_to_datetime_date(comparison_date)
        else:
            return self.formatted_date < self.__convert_to_datetime_date(comparison_date)



from datetime import date
import json
from pathlib import Path
from types import GeneratorType

import requests


# source: https://www.law.cornell.edu/uscode/text/44/3502
INDEPENDENT_REG_AGENCIES: tuple[str] = (
    'federal-reserve-system',
    'commodity-futures-trading-commission',
    'consumer-product-safety-commission',
    'federal-communications-commission',
    'federal-deposit-insurance-corporation',
    'federal-energy-regulatory-commission',
    'federal-housing-finance-agency',
    'federal-maritime-commission',
    'federal-trade-commission',
    'interstate-commerce-commission',    
    'federal-mine-safety-and-health-review-commission',
    'national-labor-relations-board',
    'nuclear-regulatory-commission',
    'occupational-safety-and-health-review-commission',
    'postal-regulatory-commission',
    'securities-and-exchange-commission',
    'consumer-financial-protection-bureau',
    'financial-research-office',
    'comptroller-of-the-currency',
    )


class AgencyMetadata:
    """Class for storing and transforming agency metadata from Federal Register API.
    
    Args:
        data (dict, optional): Accepts a JSON object of structure iterable[dict]. Defaults to None.
    """
    def __init__(self, data: list[dict] | None = None):
        if data is not None:
            self.data = data
        else:
            self.data = self.__extract_metadata()
        self.transformed_data = self.__transform()
        self.schema = self.__extract_schema()
    
    def __extract_metadata(
            self, 
            endpoint_url: str = r"https://www.federalregister.gov/api/v1/agencies.json"
        ) -> list[dict]:
        """Queries the GET agencies endpoint of the Federal Register API.
        Retrieve agencies metadata. After defining endpoint url, no parameters are needed.

        Args:
            endpoint_url (str, optional): Endpoint for retrieving agencies metadata. Defaults to r"https://www.federalregister.gov/api/v1/agencies.json".

        Raises:
            HTTPError: via requests package
        """
        # request documents; raise error if it fails
        agencies_response = requests.get(endpoint_url)
        if agencies_response.status_code != 200:
            print(agencies_response.reason)
            agencies_response.raise_for_status()
        # return response as json
        return agencies_response.json()
    
    def __extract_schema(self, metadata: dict[dict] | None = None):
        """Get Agency schema of agencies available from API.

        Args:
            metadata (dict[dict], optional): Agency metadata from API. Defaults to None.
        """        
        if metadata is not None:
            schema = [f"{slug}" for slug in metadata.get("results", {}).keys()]
        else:
            schema = [f"{agency.get('slug')}" for agency in self.data if agency.get("slug", "") != ""]
        return schema
    
    def __transform(self) -> dict[dict]:
        """Transform self.data from original format of list[dict] to dict[dict].
        """
        agency_dict = {}
        for i in self.data:
            if isinstance(i, dict):  # check if type is dict
                slug = f'{i.get("slug", "none")}'
                agency_dict.update({slug: i})                    
            else:  # cannot use this method on non-dict structures
                continue
        while "none" in agency_dict:
            agency_dict.pop("none")
        # return transformed data as a dictionary
        return agency_dict
    
    def __to_json(self, obj, path: Path, file_name: str):
        """Save object to JSON, creating path and parents if needed.

        Args:
            obj: JSON-compatible object.
            path (Path): Path for saving data.
            file_name (str): File name to use when saving.
        """        
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
        
        with open(path / file_name, "w", encoding="utf-8") as f:
            json.dump(obj, f, indent=4)
    
    def save_metadata(
            self, 
            path: Path, 
            file_name: str = "agencies_endpoint_metadata.json"
        ):
        """Save agencies metadata from Federal Register API.

        Args:
            path (Path): Path for saving data.
            file_name (str, optional): File name to use when saving. Defaults to r"agencies_endpoint_metadata.json".
        """
        # create dictionary of data with retrieval date
        dict_metadata = {
            "source": "Federal Register API, https://www.federalregister.gov/reader-aids/developer-resources/rest-api",
            "endpoint": r"https://www.federalregister.gov/api/v1/agencies.json",
            "date_retrieved": f"{date.today()}",
            "count": len(self.transformed_data), 
            "results": self.transformed_data
            }
        # export to json
        self.__to_json(dict_metadata, path, file_name)
    
    def save_schema(self, path: Path, file_name: str = "agency_schema.json"):
        """Save schema of agencies available from API.

        Args:
            path (Path): Path for saving data.
            file_name (str, optional): File name to use when saving. Defaults to "agency_schema.json".
        """        
        if (len(self.schema) == 0) and (self.data is not None):
            self.get_schema()
        self.__to_json(self.schema, path, file_name)
    
    def get_agency_metadata(self):
        """Retrieve metadata and schema from FR API GET/agencies endpoint.

        Returns:
            tuple: Transformed metadata (dict[dict]), agency schema (list[str])
        """
        return self.transformed_data, self.schema
        

class AgencyData:
    """Class for processing agency data from Federal Register API.

    Args:
        documents (list): Documents to be processed.
        metadata (dict): Transformed agency metadata from FR API.
        schema (list): Schema for valid agency slugs.
        field_keys (tuple, optional): Fields containing agency information. Defaults to ("agencies", "agency_names").
    """
    def __init__(
            self, 
            documents: list[dict], 
            metadata: dict[dict], 
            schema: list[str], 
            field_keys: tuple[str] = ("agencies", "agency_names")
        ) -> None:
            self.documents = documents
            self.field_keys = field_keys
            metadata_results = metadata.get("results", None)
            if metadata_results is not None:
                self.metadata = metadata_results
            else:
                self.metadata = metadata
            self.schema = {
                "agencies": schema, 
                "parents": self.__get_parents(), 
                "subagencies": self.__get_subagencies(), 
                }

    def __get_parents(self) -> list[str]:
        """Get top-level parent agency slugs from Agency metadata.
        Only includes agencies that have no parent themselves.
        
        Returns:
            list[str]: List of agency slugs for top-level parent agencies.
        """
        return [k for k, v in self.metadata.items() if (v.get("parent_id") is None)]
    
    def __get_subagencies(self) -> list[str]:
        """Get subagency slugs from Agency metadata.
        Includes agencies that have a parent agency, even if they are a parent themselves.
        
        Returns:
            list[str]: List of agency slugs for subagencies.
        """
        return [k for k, v in self.metadata.items() if (v.get("parent_id") is not None)]

    def __return_values_as_str(self, input_values: list | tuple | set | int | float, sep: str = "; "):
        """Return values as a string (e.g., ["a", "b", "c"] -> "a; b; c", 1.23 -> "1.23").
        Converts `list`, `tuple`, `set`, `int`, or `float` to `str`; otherwise returns value unaltered.

        Args:
            input_values (list | tuple | set | int | float): Values to convert to `str`.
            sep (str, optional): Separator for joining values. Defaults to "; ".

        Returns:
            list[str]: List of converted values for assigning to DataFrame series.
        """
        if isinstance(input_values, (list, tuple, set, GeneratorType)):
            return_value = sep.join(i for i in input_values if i is not None)
        elif isinstance(input_values, (int, float)):
            return_value = f"{input_values}"
        else:
            return_value = input_values
        return return_value
    
    def __get_agency_info(self, agency_slug: str, return_value_key: str) -> str | int | list | None:
        """Retrieve value of "return_value_key" from metadata `dict` associated with "agency_slug".

        Args:
            agency_slug (str): Agency slug identifier.
            return_value_key (str): Agency attributes associated with agency metadata (e.g., "name", "slug", "id", etc.).

        Returns:
            str | int | list | None: Value associated with "return_value_key" for given agency, else None.
        """
        return self.metadata.get(agency_slug, {}).get(return_value_key, None)
    
    def __extract_agency_slugs(self, document: dict):
        
        # 1) derive slugs from two fields
        agencies = document.get(self.field_keys[0], [])
        agency_names = document.get(self.field_keys[1], [])
        slugs = (
            agency_dict.get(
                "slug", agency_dict.get(
                    "name", f"{agency_string}").lower().replace(" ","-")
                ) for agency_dict, agency_string in zip(agencies, agency_names)
            )
        
        # 2) clean slug list to only include agencies in the schema
        # there are some bad metadata -- e.g., 'interim-rule', 'formal-comments-that-were-received-in-response-to-the-nprm-regarding'
        # also ensure no duplicate agencies in each document's list by using set()
        return list(set(slug for slug in slugs if slug in self.schema.get("agencies")))

    def __extract_parents_subagencies(
            self, 
            document: dict, 
            slug_key: str = "agency_slugs", 
            return_format: str | tuple | list | None = "slug", 
            return_values_as_str: bool = True, 
            identify_ira: bool = True
        ):
        """Extract parent and subagency information from agency data and return in requested format based on API metadata.
        Supported return formats include "child_ids", "child_slugs", "description", "id", "name", "parent_id", "short_name", "slug", "url".

        Args:
            document (dict): Federal Register document.
            slug_key (str, optional): Dictionary key containing agency slugs. Defaults to "agency_slugs".
            return_format (str, optional): Format of returned data (e.g., slug, numeric id, short name/acronym, name). Defaults to "slug".
            return_values_as_str (bool, optional): Return values as a str; otherwise returns a list. Defaults to True.
            identify_ira (bool, optional): Agency slugs contain an independent regulatory agency. Defaults to True.
        """
        if return_format is None:
            return_format = ("slug", )
        elif isinstance(return_format, str):
            return_format = (return_format, )
        elif isinstance(return_format, (tuple, list)):
            pass
        else:
            raise TypeError("Parameter 'return_format' must be `str`, `tuple`, `list`, or `None`.")
        document_copy = document.copy()
        slugs = document_copy.get(slug_key)
        if identify_ira:
                document_copy["independent_reg_agency"] = self.__identify_independent_reg_agencies(slugs)
        for fmt in return_format:
            parents = (self.__get_agency_info(slug, fmt) for slug in slugs if slug in self.schema.get("parents"))
            subagencies = (self.__get_agency_info(slug, fmt) for slug in slugs if slug in self.schema.get("subagencies"))
            if return_values_as_str:
                document_copy.update({
                    f"parent_{fmt}": self.__return_values_as_str(parents), 
                    f"subagency_{fmt}": self.__return_values_as_str(subagencies), 
                    })
            else:
                document_copy.update({
                    f"parent_{fmt}": list(parents), 
                    f"subagency_{fmt}": list(subagencies), 
                    })
        return document_copy
    
    def __identify_independent_reg_agencies(
            self, 
            slugs: list[str],
            independent_agencies: list | tuple = INDEPENDENT_REG_AGENCIES, 
            return_as_bool = True
        ):
        """Identifies whether agency slugs include an independent regulatory agency, based on the definition in [44 U.S.C. 3502(5)](https://www.law.cornell.edu/uscode/text/44/3502).
        
        Args:
            new_column (str, optional): Name of new column containing indicator for independent regulatory agencies. Defaults to "independent_reg_agency".
            independent_agencies (list | tuple, optional): Schema identifying independent regulatory agencies. Defaults to INDEPENDENT_REG_AGENCIES (constant).
        """
        ira = any(True if agency in independent_agencies else False for agency in slugs)
        if return_as_bool:
            return ira
        else:
            return 1 if ira else 0
        
    def __create_agency_slugs_key(
            self, 
            document: dict, 
            value_key: str = "agency_slugs", 
            values: list = None
        ) -> dict:
        document_copy = document.copy()
        
        document_copy.update({
            value_key: values, 
            })
        
        return document_copy
    
    def __del_field_keys(self, document: dict):
        document_copy = document.copy()
        
        for key in self.field_keys:
            document_copy.pop(key, None)
        
        return document_copy
    
    def process_data(
            self, 
            return_format: str | tuple | list | None = None, 
            return_values_as_str: bool = True, 
            identify_ira: bool = True
        ) -> list[dict]:
        """Process agency data for each document.

        Args:
            return_format (str | tuple | list | None, optional): Format of returned data (e.g., slug, numeric id, short name/acronym, name). Defaults to None.

        Returns:
            list[dict]: List of processed documents.
        """
        return [
            self.__del_field_keys(
                self.__extract_parents_subagencies(
                    self.__create_agency_slugs_key(doc, values=self.__extract_agency_slugs(doc)), 
                    return_format=return_format, 
                    return_values_as_str=return_values_as_str, 
                    identify_ira=identify_ira
                    )
                )
            for doc in self.documents
            ]


# only query agencies endpoint when run as script; save that output 
if __name__ == "__main__":
    
    # set path
    data_dir = Path(__file__).parents[1].joinpath("data")
    
    # retrieve metadata and schema
    agencies_metadata = AgencyMetadata()
    agencies_metadata.save_metadata(data_dir)
    agencies_metadata.save_schema(data_dir)



import json
from pathlib import Path

from fr_toolbelt.preprocessing import ( 
    RegInfoData, 
    )


# TEST OBJECTS AND UTILS #


TESTS_PATH = Path(__file__).parent

with open(TESTS_PATH / "test_documents.json", "r", encoding="utf-8") as f:
    TEST_DATA = json.load(f).get("results", [])


# preprocessing.rin #


def test_extract_rin_info(documents = TEST_DATA):
    rin = RegInfoData(documents)
    data = (rin._extract_field_info(doc) for doc in rin.documents)
    assert all((isinstance(d, tuple) or d is None) for d in data)


def test_create_rin_key(documents = TEST_DATA):
    rin = RegInfoData(documents)
    data = (rin._create_value_keys(doc, values=("test1", "test2")) for doc in rin.documents)
    assert all(
        isinstance(d, dict) 
        and (d.get(rin.value_keys[0]) == "test1") 
        and (d.get(rin.value_keys[1]) == "test2") 
        for d in data)


def test_process_rin_data(documents = TEST_DATA):
    rin = RegInfoData(documents)
    data = rin.process_data()
    assert isinstance(data, list)
    assert len(data) == len(documents)
